/*$Header: /home/usr/shimonyj/vto4dfp/RCS/js_nrutil.c,v 1.3 2010/02/12 05:48:36 avi Exp $*/
/*$Log: js_nrutil.c,v $
 * Revision 1.3  2010/02/12  05:48:36  avi
 * long->int to obtain 64-bit machine compatibility
 *
 * Revision 1.2  2006/12/24  06:37:00  avi
 * *** empty log message ***
 **/

#include <stdio.h>
#include <stddef.h>
#include <stdlib.h>
#include "util.h"

#define NR_END 1
#define FREE_ARG char*

/* error handler */
void nrerror(char error_text[]) 
{
	fprintf(stderr,"%s\n", error_text);
	exit(1);
}

/* allocate and free an int vector range nl..nh */
int *ivector(int nl, int nh)
{
	int *v;
	
	v = (int *)malloc( ((nh-nl+1+NR_END)*sizeof(int)));
	if (!v) nrerror("Error(ivector): allocation failure\n");
	return v-nl+NR_END;
}

void free_ivector(int *v, int nl, int nh)
{
	free((FREE_ARG) (v+nl-NR_END));
}

/* allocate and free an unsigned int vector range nl..nh */
unsigned int *lvector(int nl, int nh)
{
	unsigned int *v;
	
	v = (unsigned int *)malloc( ((nh-nl+1+NR_END)*sizeof(int)));
	if (!v) nrerror("Error(lvector): allocation failure\n");
	return v-nl+NR_END;
}

void free_lvector(unsigned int *v, int nl, int nh)
{
	free((FREE_ARG) (v+nl-NR_END));
}

/* allocate and free a float vector range nl..nh */
float *vector(int nl, int nh)
{
	float *v;
	
	v = (float *)malloc( ((nh-nl+1+NR_END)*sizeof(float)));
	if (!v) nrerror("Error(vector): allocation failure\n");
	return v-nl+NR_END;
}

void free_vector(float *v, int nl, int nh)
{
	free((FREE_ARG) (v+nl-NR_END));
}

/* allocate and free a double vector range nl..nh */
double *dvector(int nl, int nh)
{
	double *v;
	
	v = (double *)malloc( ((nh-nl+1+NR_END)*sizeof(double)));
	if (!v) nrerror("Error(dvector): allocation failure\n");
	return v-nl+NR_END;
}

void free_dvector(double *v, int nl, int nh)
{
	free((FREE_ARG) (v+nl-NR_END));
}

/* allocate and free a complex vector range nl..nh */
fcomplex *Cvector(int nl, int nh)
{
	fcomplex *v;
	
	v = (fcomplex *)malloc( ((nh-nl+1+NR_END)*sizeof(fcomplex)));
	if (!v) nrerror("Error(Cvector): allocation failure\n");
	return v-nl+NR_END;
}

void free_Cvector(fcomplex *v, int nl, int nh)
{
	free((FREE_ARG) (v+nl-NR_END));
}

/* allocate and free a int matrix with range m[nrl..nrh][ncl..nch] */
int **imatrix(int nrl, int nrh, int ncl ,int nch)
{
	int i, nrow=nrh-nrl+1, ncol=nch-ncl+1;
	int **m;
	
	/* allocate pointers to rows */
	m = (int **) malloc( ((nrow+NR_END)*sizeof(int*)));
	if (!m) nrerror("Error(imatrix): allocation failure\n");
	m += NR_END;
	m -= nrl;
	
	/* allocate rows and pointers to them */
	m[nrl] = (int *) malloc( ((nrow*ncol+NR_END)*sizeof(int)));
	if (!m[nrl]) nrerror("Error(imatrix): allocation failure\n");
	m[nrl] += NR_END;
	m[nrl] -= ncl;
	
	for (i=nrl+1; i<=nrh; i++) m[i] = m[i-1]+ncol;
	
	/* return pointer to array of pointers to rows */
	return m;
}

void free_imatrix(int **m, int nrl, int nrh, int ncl ,int nch)
{
	free((FREE_ARG) (m[nrl]+ncl-NR_END));
	free((FREE_ARG) (m+nrl-NR_END));
}


/* allocate and free a float matrix with range m[nrl..nrh][ncl..nch] */
float **matrix(int nrl, int nrh, int ncl ,int nch)
{
	int i, nrow=nrh-nrl+1, ncol=nch-ncl+1;
	float **m;
	
	/* allocate pointers to rows */
	m = (float **) malloc( ((nrow+NR_END)*sizeof(float*)));
	if (!m) nrerror("Error(matrix): allocation failure\n");
	m += NR_END;
	m -= nrl;
	
	/* allocate rows and pointers to them */
	m[nrl] = (float *) malloc( ((nrow*ncol+NR_END)*sizeof(float)));
	if (!m[nrl]) nrerror("Error(matrix): allocation failure\n");
	m[nrl] += NR_END;
	m[nrl] -= ncl;
	
	for (i=nrl+1; i<=nrh; i++) m[i] = m[i-1]+ncol;
	
	/* return pointer to array of pointers to rows */
	return m;
}

void free_matrix(float **m, int nrl, int nrh, int ncl ,int nch)
{
	free((FREE_ARG) (m[nrl]+ncl-NR_END));
	free((FREE_ARG) (m+nrl-NR_END));
}

/* allocate and free a double matrix with range m[nrl..nrh][ncl..nch] */
double **dmatrix(int nrl, int nrh, int ncl ,int nch)
{
	int i, nrow=nrh-nrl+1, ncol=nch-ncl+1;
	double **m;
	
	/* allocate pointers to rows */
	m = (double **) malloc( ((nrow+NR_END)*sizeof(double*)));
	if (!m) nrerror("Error(dmatrix): allocation failure\n");
	m += NR_END;
	m -= nrl;
	
	/* allocate rows and pointers to them */
	m[nrl] = (double *) malloc( ((nrow*ncol+NR_END)*sizeof(double)));
	if (!m[nrl]) nrerror("Error(dmatrix): allocation failure\n");
	m[nrl] += NR_END;
	m[nrl] -= ncl;
	
	for (i=nrl+1; i<=nrh; i++) m[i] = m[i-1]+ncol;
	
	/* return pointer to array of pointers to rows */
	return m;
}

void free_dmatrix(double **m, int nrl, int nrh, int ncl ,int nch)
{
	free((FREE_ARG) (m[nrl]+ncl-NR_END));
	free((FREE_ARG) (m+nrl-NR_END));
}

/* allocate and free a fcomplex matrix with range m[nrl..nrh][ncl..nch] */
fcomplex **Cmatrix(int nrl, int nrh, int ncl ,int nch)
{
	int i, nrow=nrh-nrl+1, ncol=nch-ncl+1;
	fcomplex **m;
	
	/* allocate pointers to rows */
	m = (fcomplex **) malloc( ((nrow+NR_END)*sizeof(fcomplex*)));
	if (!m) nrerror("Error(Cmatrix): allocation failure\n");
	m += NR_END;
	m -= nrl;
	
	/* allocate rows and pointers to them */
	m[nrl] = (fcomplex *) malloc( ((nrow*ncol+NR_END)*sizeof(fcomplex)));
	if (!m[nrl]) nrerror("Error(Cmatrix): allocation failure\n");
	m[nrl] += NR_END;
	m[nrl] -= ncl;
	
	for (i=nrl+1; i<=nrh; i++) m[i] = m[i-1]+ncol;
	
	/* return pointer to array of pointers to rows */
	return m;
}

void free_Cmatrix(fcomplex **m, int nrl, int nrh, int ncl ,int nch)
{
	free((FREE_ARG) (m[nrl]+ncl-NR_END));
	free((FREE_ARG) (m+nrl-NR_END));
}

